#include <vcl.h>
#include <stdio.h>
#pragma hdrstop
#include "common.h"
#include "LtrCrate.h"
#include "LtrCrateThreads.h"
#include "LtrClient.h"
#pragma package(smart_init)

#define USE_EXTENDED_PING 1

static AnsiString f_GetFPGAErrorMessage(INT fpga_err_code);

/*------------------------------------------------------------------------------------------------*/
static int f_GetCrateTypeByName(const char* id)
    {
    const struct { const char* name; int type; } list[] =
        {
        { "LTR010", CRATE_TYPE_LTR010 },
        { "LTR021", CRATE_TYPE_LTR021 },
        { "LTR030", CRATE_TYPE_LTR030 },
        { "LTR031", CRATE_TYPE_LTR031 },
        { "LTR032", CRATE_TYPE_LTR032 }
        };

    if (id) for (int i = 0; i < sizeof(list) / sizeof(list[0]); i++)
        {
        if (!strcmp(id, list[i].name))
            return list[i].type;
        }
    return CRATE_TYPE_UNKNOWN;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
TCrate::TCrate(MessageQueue<MsgControl>& mqControl, MessageQueue<MsgDisplay>& mqDisplay)
    : CurState(CRATE_ST_UNINIT),
      CrateType(CRATE_TYPE_UNKNOWN),
      protocol_version(0),
      refMQ_Control(mqControl), refMQ_Display(mqDisplay),
      ip_addr(INADDR_NONE), usb_devinst(0),
      sflags(0), baud_rate(0), hw_revision(0),
      second_mark_counter(0), start_mark_counter(0),
      unixtime_seq(0), unixtime_val(0),
      total_count_recv(0), total_count_send(0),
      special_code(false),
      ReadRawData(false), Suspended(false)
    {
    ThreadTerminationInUse = CreateMutex(NULL, FALSE, NULL);
    memset(path, 0, sizeof(path));
    memset(csn, 0, sizeof(csn));
    memset(name, 0, sizeof(name));

    hStopEvent = CreateEvent(NULL, EVT_MANUAL_RESET, EVT_NONSIGNALED, NULL);
    hClientPutEvent = CreateEvent(NULL, EVT_AUTO_RESET, EVT_NONSIGNALED, NULL);

    /*     RAW  USER  */
    Module[MODULE_INDEX_RAW].Create(TCrateModule::accReadOnly);
    Module[MODULE_INDEX_USER].Create(TCrateModule::accReadOnly);

    GetLocalTime(&start_time);
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
TCrate::~TCrate(void)
    {
    /*  ,    */
    if (WaitForSingleObject(client_list_access, INFINITE) == WAIT_OBJECT_0)
        {
        //     ,   .
        for (int k = 0; k < CLIENT_MAX; k++)
            {
            if (client_list[k] && (client_list[k]->pcrate == this))
                {
                client_list[k]->RequestTermination();
                }
            }
        for (int k = 0; k < CLIENT_MAX; k++)
            {
            if (client_list[k] && (client_list[k]->pcrate == this))
                {
                delete client_list[k]; //    
                client_list[k] = NULL;
                }
            }
        ReleaseMutex(client_list_access);
        }
    CloseHandle(hClientPutEvent);
    CloseHandle(hStopEvent);
    CloseHandle(ThreadTerminationInUse);
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
bool TCrate::ProcessCrateError(DWORD data)
    {
    BYTE regs[16];

    if (!GetArray(CM_APP, SEL_AVR_DM | 0x8000, regs, sizeof(regs)))
        {
        CancelIO();
        return false;
        }

    AnsiString msg;
    switch (data & 0xFF)
        {
        case 0:
            msg = Format("%s: fault signal REGS[0..15] =", ARRAYOFCONST((name)));
            for (int j = 0; j < sizeof(regs); j++)
                {
                msg += Format(" %0.2X", ARRAYOFCONST((regs[j])));
                }
            LOGWRITE(msg, LOGTASK_CRATE_READ, LOGL_ERROR);

            sflags |= CRATE_FLAG_HW_ERROR;
            CancelIO();
            refMQ_Display.put(LTR_DISP_crate_list_updated);
            break;
        case 1:
            { //   
            WORD mconfig_new = *(WORD*)(regs + 8);
            for (int i = 0; i < MODULE_MAX; i++)
                {
                WORD mask = (1 << i);
                if (!(mconfig_new & mask))
                    { //  ,   
                    Module[i].Delete();
                    if (baud_rate & mask)
                        { //     ,    
                        WORD baud_rate_new = baud_rate & ~mask;
                        if (PutArray(CM_APP, SEL_AVR_DM | 0x8006, (BYTE*)&baud_rate_new, 2))
                            {
                            LOGWRITE(Format("%s: change baud rate to 0x%0.4X",
                                ARRAYOFCONST((name, (int)baud_rate_new))),
                                LOGTASK_CRATE_READ, LOGL_DEBUG1);
                            }
                        else
                            {
                            LOGWRITE(Format("%s: cannot change baud rate to 0x%0.4X",
                                ARRAYOFCONST((name, (int)baud_rate_new))),
                                LOGTASK_CRATE_READ, LOGL_ERROR);
                            }
                        baud_rate = baud_rate_new;
                        }
                    }
                else
                    { // (mconfig_new & mask) == 1
                    /*           MID.
                         . */
                    Module[i].Create(TCrateModule::accNormal); //    ,   
                    }
                } /* for i (each module) */

            refMQ_Display.put(LTR_DISP_crate_list_updated);
            LOGWRITE(Format("%s: module presence mask updated to %0.4X",
                ARRAYOFCONST((name, (int)mconfig_new))), LOGTASK_CRATE_READ, LOGL_INFO);
            }
            break;
        case 2:
            LOGWRITE(Format("%s: LTR21 Buffer Overflow", ARRAYOFCONST((name))), LOGTASK_CRATE_READ, LOGL_ERROR);
            break;
        case 3:
            LOGWRITE(Format("%s: LTR21 DMA Buffer Miss", ARRAYOFCONST((name))), LOGTASK_CRATE_READ, LOGL_ERROR);
            break;
        case 4:
            LOGWRITE(Format("%s: LTR021 - Unknown Sync Source", ARRAYOFCONST((name))), LOGTASK_CRATE_READ, LOGL_ERROR);
            break;
        default:
            LOGWRITE(Format("%s: Unknown Crate Error", ARRAYOFCONST((name))), LOGTASK_CRATE_READ, LOGL_ERROR);
            break;
        } /* switch */

    return true;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
TCrateUSB::TCrateUSB(MessageQueue<MsgControl>& mqControl, MessageQueue<MsgDisplay>& mqDisplay)
    : TCrate(mqControl, mqDisplay)
    {
    recv_buf_thread = NULL;
    send_buf_thread = NULL;
    DeviceHANDLE = NULL;
    ltr_USB = CreateInstanceUSB();
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
TCrateIP::TCrateIP(MessageQueue<MsgControl>& mqControl, MessageQueue<MsgDisplay>& mqDisplay, DWORD IdleTimeout)
    : TCrate(mqControl, mqDisplay)
    {
    ping_fail_count = 0;
    data_thread = NULL;
    ltr_ip = CreateInstanceIP(hStopEvent, IdleTimeout);
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
TCrateUSB::~TCrateUSB(void)
    {
    CurState = CRATE_ST_DEAD;
    StopServiceThreads();
    for (int i = 0; i < MODULE_ARRAYLEN; i++)
        {
        Module[i].Delete();
        }
    if (ltr_USB != NULL)
        DeleteInstanceUSB(ltr_USB);
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
TCrateIP::~TCrateIP(void)
    {
    CurState = CRATE_ST_DEAD;
    StopServiceThreads();
    for (int i = 0; i < MODULE_ARRAYLEN; i++)
        {
        Module[i].Delete();
        }
    if (ltr_ip != NULL)
        DeleteInstanceIP(ltr_ip);
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
AnsiString TCrateUSB::IoRetryStr(void)
    {
    unsigned int retries = ltr_USB->GetLastIoctlRetries();
    if (!retries)
        return "";
    return Format(" (retries: %d)", ARRAYOFCONST((retries)));
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
void TCrateUSB::SetState(t_CrateState new_state, bool NotifyThreads)
    {
    t_CrateState old_state = CurState;
    CurState = new_state;
    if (NotifyThreads && (new_state != old_state))
        {
        /*    ONLINE  -ONLINE,      */
        if ((new_state == CRATE_ST_ONLINE) || (old_state == CRATE_ST_ONLINE))
            refMQ_Display.put(LTR_DISP_crate_list_updated);
        /*     DEAD    */
        if (new_state == CRATE_ST_DEAD)
            refMQ_Control.put(LTR_SRV_bury_the_dead);
        }
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
void TCrateIP::SetState(t_CrateState new_state, bool NotifyThreads)
    {
    t_CrateState old_state = CurState;
    CurState = new_state;
    switch (new_state)
        {
        case CRATE_ST_ONLINE:
            IPCrateList.SetStatus(ip_addr, IP_STATUS_ONLINE, csn);
            break;
        case CRATE_ST_CONNECTING:
            IPCrateList.SetStatus(ip_addr, IP_STATUS_CONNECTING);
            break;
        default:
            IPCrateList.SetStatus(ip_addr, IP_STATUS_OFFLINE);
            break;
        }
    if (NotifyThreads && (new_state != old_state))
        {
        /*    ONLINE  -ONLINE,      */
        if ((new_state == CRATE_ST_ONLINE) || (old_state == CRATE_ST_ONLINE))
            refMQ_Display.put(LTR_DISP_crate_list_updated);
        /*     DEAD    */
        if (new_state == CRATE_ST_DEAD)
            refMQ_Control.put(LTR_SRV_bury_the_dead);
        /*  IP       IP */
        refMQ_Display.put(LTR_DISP_ip_list_updated);
        }
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
BYTE TCrateUSB::GetConnType(void)
    {
    return CRATE_IFACE_USB;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
BYTE TCrateIP::GetConnType(void)
    {
    return CRATE_IFACE_TCPIP;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
void TCrateUSB::SuspendCrate(void)
    {
    if (Suspended) return;
    Suspended = true;

    StopServiceThreads();

    ltr_USB->CancelIO();
    if (ltr_USB->CloseLDevice())
        {
        LOGWRITE(Format("%s: crate suspended", ARRAYOFCONST((name))),
            LOGTASK_CRATE_CTL, LOGL_INFO);
        return;
        }
    else
        {
        Suspended = false;
        LOGWRITE(Format("%s: cannot suspend crate", ARRAYOFCONST((name))),
            LOGTASK_CRATE_CTL, LOGL_ERROR);
        }
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
void TCrateIP::SuspendCrate(void)
    {
    if (Suspended) return;
    Suspended = true;
    StopServiceThreads();
    ltr_ip->StopIdleTimer();
    LOGWRITE(Format("%s: crate suspended", ARRAYOFCONST((name))),
        LOGTASK_CRATE_CTL, LOGL_INFO);
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
void TCrateUSB::ResumeCrate(void)
    {
    if (!Suspended) return;
    if (ltr_USB->OpenLDeviceByName((char*)path))
        {
        LOGWRITE(Format("%s: crate resumed from suspend", ARRAYOFCONST((name))) + IoRetryStr(),
            LOGTASK_CRATE_CTL, LOGL_INFO);
        Suspended = false;
        StartServiceThreads();
        return;
        }
    LOGWRITE(Format("%s: cannot resume crate from suspend", ARRAYOFCONST((name))),
        LOGTASK_CRATE_CTL, LOGL_ERROR);
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
void TCrateIP::ResumeCrate(void)
    {
    if (!Suspended) return;
    ltr_ip->RestartIdleTimer();
    if (StartServiceThreads())
        {
        LOGWRITE(Format("%s: crate resumed from suspend", ARRAYOFCONST((name))),
            LOGTASK_CRATE_CTL, LOGL_INFO);
        Suspended = false;
        return;
        }
    LOGWRITE(Format("%s: cannot resume crate from suspend", ARRAYOFCONST((name))),
        LOGTASK_CRATE_CTL, LOGL_ERROR);
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
bool TCrateUSB::Startup(BYTE * FPGAData, bool use_default_ttf)
    { /*  USB   ,    */
    if (GetState() != CRATE_ST_UNINIT)
        { /*    */
        StopServiceThreads();
        sflags &= ~CRATE_FLAG_BOOT_LOADER;
        }
    STRNCPYZ(name, "Init USB crate");
    SetState(CRATE_ST_CONNECTING);
    bool result = InitCrate(FPGAData, use_default_ttf);
    if (result)
        {
        /*     InitCrate() */
        SetState(CRATE_ST_ONLINE);
        }
    else
        {
        SetState(CRATE_ST_DEAD);
        }
    return result;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
bool TCrateIP::Startup(BYTE * FPGAData, bool use_default_ttf)
    { /*  TCP      */
    if (GetState() != CRATE_ST_UNINIT)
        { /*    */
        StopServiceThreads();
        sflags &= ~CRATE_FLAG_BOOT_LOADER;
        }
    //      
    snprintf(name, sizeof(name), "[%s]", path);
    SetState(CRATE_ST_CONNECTING);
    if (!StartServiceThreads())
        {
        LOGWRITE(Format("%s: cannot start data thread", ARRAYOFCONST((name))),
            LOGTASK_CRATE_INIT, LOGL_ERROR);
        SetState(CRATE_ST_DEAD);
        return false;
        }

    return true; /*   --     */
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
bool TCrateUSB::InitCrate(BYTE * FPGAData, bool use_default_ttf)
    {
//         TLTRServiceCrate::ConnectUSBCrate()
//    LOGWRITE("Connecting USB crate...", LOGTASK_CRATE_INIT, LOGL_INFO);

    baud_rate = 0;

    if (!ltr_USB->OpenLDeviceByName((char*)path))
        {
        LOGWRITE(Format("%s: cannot open device %s", ARRAYOFCONST((name, path))),
            LOGTASK_CRATE_INIT, LOGL_ERROR);
        return false;
        }

    LOGWRITE(Format("%s: OpenLDeviceByName() ok",  ARRAYOFCONST((name))) + IoRetryStr(),
        LOGTASK_CRATE_INIT, LOGL_DEBUG3);

    //  bootloader  aplication
    char appmode = '?';
    if (ltr_USB->SwitchBootloader(true))
        {
        appmode = 'B'; // bootloader active
        }
    else
        {
        LOGWRITE(Format("%s: bootloader mode select failed (error %d)",
            ARRAYOFCONST((name, (int)GetLastError()))), LOGTASK_CRATE_INIT, LOGL_DEBUG2);
        }
    if (ltr_USB->SwitchBootloader(false))
        {
        appmode = 'A'; // application active
        }
    else
        {
        LOGWRITE(Format("%s: application mode select failed (error %d)",
            ARRAYOFCONST((name, (int)GetLastError()))), LOGTASK_CRATE_INIT, LOGL_DEBUG2);
        }
    if (appmode == '?')
        {
        LOGWRITE(Format("%s: cannot set application mode", ARRAYOFCONST((name))),
            LOGTASK_CRATE_INIT, LOGL_ERROR);
        return false;
        }

    TDESCRIPTION_LTR010 descr;
    char CrateName[16];

    if (!ltr_USB->GetModuleName(CrateName, sizeof(CrateName)))
        {
        LOGWRITE(Format("%s: error in GetModuleName", ARRAYOFCONST((name))),
            LOGTASK_CRATE_INIT, LOGL_ERROR);
        return false;
        }
    LOGWRITE(Format("%s: GetModuleName() ok",  ARRAYOFCONST((name))) + IoRetryStr(),
        LOGTASK_CRATE_INIT, LOGL_DEBUG3);

    if (!GetModuleDescription(CM_AUTO, &descr))
        {
        LOGWRITE(Format("%s: error in GetModuleDescription", ARRAYOFCONST((name))),
            LOGTASK_CRATE_INIT, LOGL_ERROR);
        return false;
        }
    LOGWRITE(Format("%s: GetModuleDescription() ok", ARRAYOFCONST((name))) + IoRetryStr(),
        LOGTASK_CRATE_INIT, LOGL_DEBUG3);

    ZeroMemory(csn, sizeof(csn));
    memcpy(csn, descr.Module.SerialNumber,
        std::min(sizeof(csn) - 1, sizeof(descr.Module.SerialNumber)));

    //      
    descr.Module.DeviceName[sizeof(descr.Module.DeviceName) - 1] = '\0';
    snprintf(name, sizeof(name), "%s [%s]", descr.Module.DeviceName, csn);

    LOGWRITE("Crate Info:", LOGTASK_CRATE_INIT, LOGL_INFO);
    LOGWRITE(Format("+ DeviceName = %.16s", ARRAYOFCONST(((char*)descr.Module.DeviceName))), LOGTASK_CRATE_INIT, LOGL_INFO);
    LOGWRITE(Format("+ SerialNumber = %.16s", ARRAYOFCONST(((char*)descr.Module.SerialNumber))), LOGTASK_CRATE_INIT, LOGL_INFO);
    LOGWRITE(Format("+ DeviceComment = %s", ARRAYOFCONST(((char*)descr.Module.Comment))), LOGTASK_CRATE_INIT, LOGL_VERBOSE);
    LOGWRITE("+ Connection = " + USB_ConnInfoString(usb_info), LOGTASK_CRATE_INIT, LOGL_VERBOSE);

    int v1 = 0, v2 = 0, v3 = 0, v4 = 0;
    if (sscanf(descr.Cpu.Comment, "Bootloader v%d.%d.%d.%d", &v1, &v2, &v3, &v4)==4)
        {
        LOGWRITE(Format("+ BootLoaderVersion = %d.%d.%d.%d", ARRAYOFCONST((v1, v2, v3, v4))), LOGTASK_CRATE_INIT, LOGL_VERBOSE);
        }
    if (descr.Cpu.FirmwareVersion!=0xFFFFFFFF)
        {
        LOGWRITE(Format("+ FirmwareVersion = %d.%d.%d.%d", ARRAYOFCONST((
            (int)(descr.Cpu.FirmwareVersion>>24)&0xFF,
            (int)(descr.Cpu.FirmwareVersion>>16)&0xFF,
            (int)(descr.Cpu.FirmwareVersion>> 8)&0xFF,
            (int)(descr.Cpu.FirmwareVersion>> 0)&0xFF))),
            LOGTASK_CRATE_INIT, LOGL_VERBOSE);
        }

    //     -
    AnsiString crate_ttf = "";

    if (use_default_ttf)
        {
        crate_ttf = g_Conf.GetTTFFileName(
            Format("LTR010_%.16s", ARRAYOFCONST((csn)))
            );
        }

    if (appmode == 'A')
        {
        // application
        CrateType = f_GetCrateTypeByName(CrateName);

        switch (CrateType)
            {
            case CRATE_TYPE_LTR030:
            case CRATE_TYPE_LTR031:
            case CRATE_TYPE_LTR032:
                if (ltr_USB->Application->GetArray(SEL_DSP | 0x8001, &hw_revision, sizeof(hw_revision)))
                    {
                    LOGWRITE("+ BoardRevision = " + IntToStr(hw_revision) + IoRetryStr(),
                        LOGTASK_CRATE_INIT, LOGL_VERBOSE);
                    }
                else
                    {
                    hw_revision = 0;
                    }
                // no break!
            case CRATE_TYPE_LTR010:
            case CRATE_TYPE_LTR021:
                {
                if (FPGAData || !crate_ttf.IsEmpty())
                    {
                    bool fpga_ok;
                    INT err_code;
                    AnsiString from_file;
                    if (FPGAData)
                        {
                        fpga_ok = ltr_USB->Application->LoadFPGA(FPGAData, 7, 0, &err_code);
                        }
                    else
                        {
                        from_file = " from " + crate_ttf;
                        fpga_ok = ltr_USB->Application->LoadFPGAFromFile(crate_ttf.c_str(), 7, 0, &err_code);
                        }

                    if (fpga_ok)
                        {
                        LOGWRITE("LoadFPGA" + from_file + ": ok" +
                            ((err_code) ? Format(" (retries: %d)", ARRAYOFCONST((err_code))) : AnsiString()),
                            LOGTASK_CRATE_INIT, LOGL_VERBOSE);
                        }
                    else
                        {
                        /* @@FIXME --     LTR21 */
                        bool non_fatal = (CRATE_TYPE_LTR021 == CrateType) && (LTR_FPGA_ERR_NOREPLY == err_code);
                        LOGWRITE(Format("%s: cannot load FPGA" + from_file +": ", ARRAYOFCONST((name))) +
                            f_GetFPGAErrorMessage(err_code),
                            LOGTASK_CRATE_INIT, (non_fatal) ? LOGL_WARNING : LOGL_ERROR);
                        if (!non_fatal)
                            return false;
                        }
                    }

                ltr_USB->Application->ConvertFPGADescription(&descr);
                LOGWRITE(Format("+ FpgaName = %s", ARRAYOFCONST(((char*)descr.Fpga.Name))), LOGTASK_CRATE_INIT, LOGL_VERBOSE);
                LOGWRITE(Format("+ FpgaVersion = %d.%d.%d.%d", ARRAYOFCONST((
                    (int)(descr.Fpga.FirmwareVersion>>24)&0xFF,
                    (int)(descr.Fpga.FirmwareVersion>>16)&0xFF,
                    (int)(descr.Fpga.FirmwareVersion>> 8)&0xFF,
                    (int)(descr.Fpga.FirmwareVersion>> 0)&0xFF))),
                    LOGTASK_CRATE_INIT, LOGL_VERBOSE);
                LOGWRITE(Format("+ FpgaComment = %s", ARRAYOFCONST(((char*)descr.Fpga.Comment))),
                    LOGTASK_CRATE_INIT, LOGL_VERBOSE);
                }
                break;
            } /* Switch CrateType */

        if (ltr_USB->Application->PutArray(SEL_AVR_DM | 0x8006, (BYTE*)&baud_rate, 2))
            {
            LOGWRITE(Format("+ BaudRate = 0x%0.4X", ARRAYOFCONST(((int)baud_rate))) + IoRetryStr(),
                LOGTASK_CRATE_INIT, LOGL_VERBOSE);
            }
        else
            {
            LOGWRITE(Format("%s: cannot set baud rate after Load FPGA",
                ARRAYOFCONST((name))), LOGTASK_CRATE_INIT, LOGL_ERROR);
            return false;
            }

        if (ltr_USB->Application->GetModuleDescription(&descr))
            {
            LOGWRITE("+ GetModuleDescription() ok" + IoRetryStr(),
                LOGTASK_CRATE_INIT, LOGL_DEBUG3);
            }
        else
            {
            LOGWRITE(Format("%s: error in GetModuleDescription",
                ARRAYOFCONST((name))), LOGTASK_CRATE_INIT, LOGL_ERROR);
            return false;
            }
        } /* appmode == 'A' */
    else
        {
        // bootloader
        CrateType = CRATE_TYPE_BOOTLOADER;
        sflags |= CRATE_FLAG_BOOT_LOADER;
        }

    /*   - */
    if (!StartServiceThreads())
        {
        LOGWRITE(Format("%s: cannot start crate service threads", ARRAYOFCONST((name))),
            LOGTASK_CRATE_INIT, LOGL_ERROR);
        return false;
        }

    return true;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
bool TCrateIP::InitCrate(void)
    {
    baud_rate = 0;
    LOGWRITE(Format("Connecting IP crate %s ...", ARRAYOFCONST((path))),
        LOGTASK_CRATE_INIT, LOGL_INFO);
    if (!ltr_ip->OpenDeviceByIP(ip_addr))
        {
        LOGWRITE(Format("%s: cannot connect to %s: %s",
            ARRAYOFCONST((name, path, GetErrorMessage()))),
            LOGTASK_CRATE_INIT, LOGL_ERROR);
        return false;
        }

    ltr_ip->SetTimeouts(g_Conf.CrateConnectTimeout(), g_Conf.CrateCommandTimeout());

    protocol_version = ltr_ip->GetProtocolVersion();
    bool supports_mcs = false;

    LOGWRITE(Format("%s: OpenLDeviceByIP() ok, protocol version %d.%d",
        ARRAYOFCONST((name, protocol_version >> 8, protocol_version & 0xFF))),
        LOGTASK_CRATE_INIT, LOGL_DEBUG3);

    /*      1.1+ */
    if (protocol_version >= 0x0101)
        {
        BYTE is_primary_conn;
        /*        ,  @@ ?@@
                 ,   */
        if (!ltr_ip->GetArray(SEL_AVR_DM | 0x10070, &is_primary_conn, sizeof(is_primary_conn)))
            {
            LOGWRITE(Format("%s: cannot check for primary connection: %s",
                ARRAYOFCONST((name, GetErrorMessage()))),
                LOGTASK_CRATE_INIT, LOGL_ERROR);
            return false;
            }

        if (!is_primary_conn)
            { 
            //@@ test -- always force
            LOGWRITE(Format("%s: crate reported already open, forcing connection",
                ARRAYOFCONST((name))), LOGTASK_CRATE_INIT, LOGL_WARNING);
            if (!ltr_ip->PutArray(SEL_AVR_DM | 0x10072, NULL, 0))
                {
                LOGWRITE(Format("%s: cannot force primary connection: %s",
                    ARRAYOFCONST((name, GetErrorMessage()))),
                    LOGTASK_CRATE_INIT, LOGL_ERROR);
                return false;
                }
            }

        /*    module control structure */
        WORD mcs_ver;
        if (!ltr_ip->GetArray(SEL_AVR_DM | 0x10101, (BYTE*)&mcs_ver, sizeof(mcs_ver)))
            {
            LOGWRITE(Format("%s: cannot query MCS capability: %s",
                ARRAYOFCONST((name, GetErrorMessage()))),
                LOGTASK_CRATE_INIT, LOGL_ERROR);
            return false;
            }
        supports_mcs = (mcs_ver > 0);
        if (supports_mcs)
            {
            LOGWRITE(Format("%s: crate has MCS capability (version %d)",
                ARRAYOFCONST((name, mcs_ver))), LOGTASK_CRATE_INIT, LOGL_DEBUG1);
            }
        else
            {
            LOGWRITE(Format("%s: crate has no MCS capability",
                ARRAYOFCONST((name))), LOGTASK_CRATE_INIT, LOGL_DEBUG1);
            }

        } /* if (protocol_version >= 0x0101) */

    TDESCRIPTION_LTR010 descr;
    char CrateName[16];

    if (!ltr_ip->GetModuleName(CrateName, sizeof(CrateName)))
        {
        LOGWRITE(Format("%s: error in GetModuleName: %s",
            ARRAYOFCONST((name, GetErrorMessage()))),
            LOGTASK_CRATE_INIT, LOGL_ERROR);
        return false;
        }
    LOGWRITE(Format("%s: GetModuleName() ok", ARRAYOFCONST((name))), LOGTASK_CRATE_INIT, LOGL_DEBUG3);

    if (!ltr_ip->GetModuleDescription(&descr))
        {
        LOGWRITE(Format("%s: error in GetModuleDescription: %s",
            ARRAYOFCONST((name, GetErrorMessage()))),
            LOGTASK_CRATE_INIT, LOGL_ERROR);
        return false;
        }
    LOGWRITE(Format("%s: GetModuleDescription() ok", ARRAYOFCONST((name))), LOGTASK_CRATE_INIT, LOGL_DEBUG3);

    ZeroMemory(csn, sizeof(csn));
    memcpy(csn, descr.Module.SerialNumber,
        std::min(sizeof(csn) - 1, sizeof(descr.Module.SerialNumber)));

    //      
    descr.Module.DeviceName[sizeof(descr.Module.DeviceName) - 1] = '\0';
    snprintf(name, sizeof(name), "%s [%s]", descr.Module.DeviceName, csn);

    LOGWRITE("Crate Info:", LOGTASK_CRATE_INIT, LOGL_INFO);
    LOGWRITE(Format("+ Location = %s", ARRAYOFCONST(((char*)path))), LOGTASK_CRATE_INIT, LOGL_INFO);
    LOGWRITE(Format("+ DeviceName = %.16s", ARRAYOFCONST(((char*)descr.Module.DeviceName))), LOGTASK_CRATE_INIT, LOGL_INFO);
    LOGWRITE(Format("+ SerialNumber = %.16s", ARRAYOFCONST(((char*)descr.Module.SerialNumber))), LOGTASK_CRATE_INIT, LOGL_INFO);
    LOGWRITE(Format("+ DeviceComment = %s", ARRAYOFCONST(((char*)descr.Module.Comment))), LOGTASK_CRATE_INIT, LOGL_VERBOSE);

    int v1 = 0, v2 = 0, v3 = 0, v4 = 0;
    if (sscanf(descr.Cpu.Comment, "Bootloader v%d.%d.%d.%d", &v1, &v2, &v3, &v4) == 4)
        {
        LOGWRITE(Format("+ BootLoaderVersion = %d.%d.%d.%d", ARRAYOFCONST((v1, v2, v3, v4))),
            LOGTASK_CRATE_INIT, LOGL_VERBOSE);
        }

    if (descr.Cpu.FirmwareVersion != 0xFFFFFFFF)
        {
        LOGWRITE(Format("+ FirmwareVersion = %d.%d.%d.%d", ARRAYOFCONST((
            (int)(descr.Cpu.FirmwareVersion >> 24) & 0xFF,
            (int)(descr.Cpu.FirmwareVersion >> 16) & 0xFF,
            (int)(descr.Cpu.FirmwareVersion >> 8) & 0xFF,
            (int)(descr.Cpu.FirmwareVersion) & 0xFF))),
            LOGTASK_CRATE_INIT, LOGL_VERBOSE);
        }

    if (ltr_ip->GetArray(SEL_DSP | 0x8001, &hw_revision, sizeof(hw_revision)))
        {
        LOGWRITE("+ BoardRevision = " + IntToStr(hw_revision),
            LOGTASK_CRATE_INIT, LOGL_VERBOSE);
        }
    else
        {
        hw_revision = 0;
        }

    LOGWRITE(Format("+ FpgaName = %s", ARRAYOFCONST(((char*)descr.Fpga.Name))),
        LOGTASK_CRATE_INIT, LOGL_VERBOSE);
    LOGWRITE(Format("+ FpgaVersion = %d.%d.%d.%d", ARRAYOFCONST((
        (int)(descr.Fpga.FirmwareVersion >> 24) & 0xFF,
        (int)(descr.Fpga.FirmwareVersion >> 16) & 0xFF,
        (int)(descr.Fpga.FirmwareVersion >> 8) & 0xFF,
        (int)(descr.Fpga.FirmwareVersion) & 0xFF))),
        LOGTASK_CRATE_INIT, LOGL_VERBOSE);
    LOGWRITE(Format("+ FpgaComment = %s", ARRAYOFCONST(((char*)descr.Fpga.Comment))),
        LOGTASK_CRATE_INIT, LOGL_VERBOSE);

    CrateType = f_GetCrateTypeByName(CrateName);

    /*    ini- */
    //AnsiString section = Format("LTR010_%.16s", ARRAYOFCONST(((char*)descr.Module.SerialNumber)));

    /*  FPGA   --  ,  LAN,  .   fpga_info
           . */

    /* Set baud rate */
    if (ltr_ip->PutArray(SEL_AVR_DM | 0x8006, (BYTE*)&baud_rate, 2))
        {
        LOGWRITE(Format("+ BaudRate = 0x%0.4X", ARRAYOFCONST(((int)baud_rate))),
            LOGTASK_CRATE_INIT, LOGL_VERBOSE);
        }
    else
        {
        LOGWRITE(Format("%s: cannot set baudrate: ",
            ARRAYOFCONST((name, GetErrorMessage()))),
            LOGTASK_CRATE_INIT, LOGL_ERROR);
        return false;
        }

    /*  Module Control Structure:     
          STOP-RESET-STOP,     */
    if (supports_mcs)
        {
        BYTE regs[16];
        /*      */
        if (!ltr_ip->GetArray(SEL_AVR_DM | 0x8000, regs, sizeof(regs)))
            {
            LOGWRITE(Format("%s: cannot query module presence: %s",
                ARRAYOFCONST((name, GetErrorMessage()))),
                LOGTASK_CRATE_INIT, LOGL_ERROR);
            return false;
            }

        WORD presence_mask = *(WORD*)(regs + 8);
        WORD mcs_mid[MODULE_MAX];

        LOGWRITE(Format("%s: module presence mask = %0.4X",
            ARRAYOFCONST((name, (int)presence_mask))), LOGTASK_CRATE_INIT, LOGL_INFO);

        for (int i = 0; i < MODULE_MAX; i++)
            {
            mcs_mid[i] = 0;
            if (presence_mask & (1 << i))
                {
                DWORD mcs_header;
                Module[i].Create(TCrateModule::accNormal);
                if (ltr_ip->GetArray((SEL_AVR_DM | 0x10060) + i,
                    (BYTE*)&mcs_header, sizeof(mcs_header)))
                    {
                    const char* const mcshdr_str_v[2] = { "INVALID", "VALID" };
                    const char* const mcshdr_str_t[4] = { "FREE", "INIT", "KEEP", "???" };
                    int v = (mcs_header >> 30) & 1;
                    int t = (mcs_header >> 28) & 3;
                    int mcs_size = (mcs_header >> 8) & 0xFFFFF;
                    int mid = mcs_header & 0xFF;
                    LOGWRITE(Format("%s: [MCS slot %d] header = %0.8X (v = %s, t = %s, size = %d, mid = %d)",
                        ARRAYOFCONST((name, 1 + i, mcs_header, mcshdr_str_v[v], mcshdr_str_t[t],
                        mcs_size, mid))), LOGTASK_CRATE_INIT, LOGL_DEBUG3);
                    if (v && t)
                        {
                        mcs_mid[i] = mid | (mid << 8);
                        LOGWRITE(Format("%s: [MCS slot %d] picking up LTR%d without reset",
                            ARRAYOFCONST((name, 1 + i, mid))), LOGTASK_CRATE_INIT, LOGL_INFO);
                        }
                    }
                else
                    {
                    LOGWRITE(Format("%s: [MCS slot %d] cannot read header: %s",
                        ARRAYOFCONST((name, 1 + i, GetErrorMessage()))), LOGTASK_CRATE_INIT, LOGL_WARNING);
                    }
                }
            } /* for i */

        LockAllModules();
        for (int i = 0; i < MODULE_MAX; i++)
            {
            if (mcs_mid[i])
                { /*    ,    stop-reset-stop */
                InitLockedModule(i, mcs_mid[i], module_init_no_reset);
                }
            } /* for i */
        UnlockAllModules();

        refMQ_Display.put(LTR_DISP_crate_list_updated);
        }

    return true;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
bool TCrateUSB::StartServiceThreads(void)
    {
    StopServiceThreads();

    DeviceHANDLE = ltr_USB->GetUSBDeviceHandle();

    if (!DeviceHANDLE)
        {
        LOGWRITE(Format("%s: cannot start service threads on unconnected device",
            ARRAYOFCONST((name))),
            LOGTASK_CRATE_INIT, LOGL_ERROR);
        return false;
        }

    ResetEvent(hStopEvent);

    recv_buf_thread = new TReadUSBCrateThread(this, false);
    if (WaitForSingleObject(((TReadUSBCrateThread*)recv_buf_thread)->init_complete, INFINITE) != WAIT_OBJECT_0)
        {
        LOGWRITE(Format("%s: cannot init read thread", ARRAYOFCONST((name))), LOGTASK_CRATE_INIT, LOGL_ERROR);
        StopServiceThreads();
        return false;
        }

    send_buf_thread = new TWriteUSBCrateThread(this, false);
    if (WaitForSingleObject(((TWriteUSBCrateThread*)send_buf_thread)->init_complete, INFINITE) != WAIT_OBJECT_0)
        {
        LOGWRITE(Format("%s: cannot init write thread", ARRAYOFCONST((name))), LOGTASK_CRATE_INIT, LOGL_ERROR);
        StopServiceThreads();
        return false;
        }

    LOGWRITE(Format("%s: crate service started.", ARRAYOFCONST((name))), LOGTASK_CRATE_INIT, LOGL_DEBUG1);
    return true;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
bool TCrateIP::StartServiceThreads(void)
    {
    StopServiceThreads();

    ResetEvent(hStopEvent);

    data_thread = new TIPCrateDataThread(this, false);
    if (WaitForSingleObject(((TIPCrateDataThread*)data_thread)->init_complete, INFINITE) != WAIT_OBJECT_0)
        {
        LOGWRITE(Format("%s: cannot init data thread", ARRAYOFCONST((name))), LOGTASK_CRATE_INIT, LOGL_ERROR);
        StopServiceThreads();
        return false;
        }

    LOGWRITE(Format("%s: crate service started.", ARRAYOFCONST((name))), LOGTASK_CRATE_INIT, LOGL_DEBUG1);
    return true;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
void TCrateUSB::StopServiceThreads()
    {
    WaitForSingleObject(ThreadTerminationInUse, INFINITE);

    ltr_USB->CancelIO();

    SetEvent(hStopEvent);

    if (recv_buf_thread)
        recv_buf_thread->Terminate();

    if (send_buf_thread)
        send_buf_thread->Terminate();

    StopThread(recv_buf_thread);
    StopThread(send_buf_thread);

    DeviceHANDLE = NULL;

    ReleaseMutex(ThreadTerminationInUse);
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
void TCrateIP::StopServiceThreads()
    {
    WaitForSingleObject(ThreadTerminationInUse, INFINITE);

    SetEvent(hStopEvent);

    StopThread(data_thread);

    ReleaseMutex(ThreadTerminationInUse);
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
void __fastcall TCrate::ProcessSyncMark(WORD mark_val)
    {
    /* old mark: 00TS xxxx xxxx xxxx
       new mark: 01TS tttt tttt tttt = unixtime[11:0]
                 10TS tttt tttt tttt = unixtime[23:12]
                 11TS tttt tttt tttt = unixtime[35:24]
       T = second
       S = start
     */
    enum en_sync_mark_const
        {
        MARK_FLAG_START = 0x1000,
        MARK_FLAG_SECOND = 0x2000
        };

    DWORD out_word[4];
    int out_count = 0;

    if (mark_val & MARK_FLAG_START) start_mark_counter++;
    if (mark_val & MARK_FLAG_SECOND) second_mark_counter++;

    DWORD tmark = ((start_mark_counter & 0xFFFF) << 16) | (second_mark_counter & 0xFFFF);

    switch (mark_val >> 14)
        {
        case 0: //   
            if (mark_val & (MARK_FLAG_START | MARK_FLAG_SECOND))
                {
                out_word[out_count++] = LTR_SIGNAL_TSTAMP;
                out_word[out_count++] = tmark;
                }
            break;
        case 1: //    (  unixtime)
            unixtime_seq = 1;
            unixtime_val = mark_val & 0xFFF;
            break;
        case 2: //    (  unixtime)
            if (++unixtime_seq == 2)
                {
                unixtime_val |= (__int64)(mark_val & 0xFFF) << 12;
                }
            else
                { //  
                unixtime_seq = 0;
                }
            break;
        case 3: //    (  unixtime)
            if (++unixtime_seq == 3)
                {
                unixtime_val |= (__int64)(mark_val & 0xFFF) << 24;
                if (g_Conf.LongTStampEnabled())
                    {
                    out_word[out_count++] = LTR_SIGNAL(LTR_SIG_TYPE_TSTAMP_EX, 3);
                    out_word[out_count++] = tmark;
                    out_word[out_count++] = (DWORD)unixtime_val;
                    out_word[out_count++] = (DWORD)(unixtime_val >> 32);
                    }
                else
                    {
                    out_word[out_count++] = LTR_SIGNAL_TSTAMP;
                    out_word[out_count++] = tmark;
                    }
                }
            else
                { //  
                unixtime_seq = 0;
                }
            break;
        }

    if (out_count < 1)
        return;

    for (int k = 0; k < MODULE_MAX; k++)
        {
        for (int i = 0; i < out_count; i++)
            Module[k].pvt_CratePut(out_word[i]);
        }

    if (ReadRawData)
        {
        for (int i = 0; i < out_count; i++)
            Module[MODULE_INDEX_RAW].pvt_CratePut(out_word[i]);
        }
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
void TCrate::InitLockedModule(int ModuleSlot, WORD new_mid, TCrate::module_init_mode_t init_mode)
    {
    bool is_fast = false;

    Module[ModuleSlot].pvt_InitMID(new_mid);
    if (module_init_no_reset == init_mode)
        Module[ModuleSlot].pvt_GetAndClearInitFlag(); //  stop-reset-stop

    switch (new_mid)
        {
        //  
        case LTR_MID(1): // LTR01 -   
        case LTR_MID(24):
        case LTR_MID(34):
        case LTR_MID(35):
            is_fast = true;
            break;

        //   
        case LTR_MID(41):
        case LTR_MID(42):
        case LTR_MID(43):
        case LTR_MID(21): // LTR21 -  -
            start_mark_counter = second_mark_counter = 0; //  
            break;
        default:
            break;
        } /* switch (new_mid) */

    LOGWRITE(Format("%s: found new module LTR%d (slot = %d, mid = %0.4X%s)",
        ARRAYOFCONST((name, (int)new_mid & 0xFF, (int)ModuleSlot + 1, (int)new_mid,
        ((is_fast) ? ", fast" : "")))),
        LOGTASK_CRATE_READ, LOGL_INFO);

    /* 20-Sep-2012 :  (?)    mid (Module[ModuleSlot].mid) */
    if (is_fast /* && (mid == 0xFFFF) */) 
        {
        WORD baud_rate_new = baud_rate | (1 << ModuleSlot);
        if (PutArray(CM_APP, SEL_AVR_DM | 0x8006, (BYTE*)&baud_rate_new, 2))
            {
            LOGWRITE(Format("%s: change baud rate to 0x%0.4X",
                ARRAYOFCONST((name, (int)baud_rate_new))),
                LOGTASK_CRATE_READ, LOGL_DEBUG1);
            }
        else
            {
            LOGWRITE(Format("%s: cannot change baud rate to 0x%0.4X",
                ARRAYOFCONST((name, (int)baud_rate_new))),
                LOGTASK_CRATE_READ, LOGL_ERROR);
            }
        baud_rate = baud_rate_new;
        }
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
bool TCrate::ParseDataFromCrate(LPDWORD lpBuffer, DWORD NumDWords)
    {
    bool result = true;

    if ((NumDWords == 0) || (lpBuffer == NULL))
        return false;

    DEBUGDUMP(lpBuffer, NumDWords * sizeof(DWORD), Format("from crate %s", ARRAYOFCONST((name))).c_str());
#if DEBUG_PROFILER
    Prf::Add _(1);
#endif

    LockAllModules();

    //  
    total_count_recv += NumDWords;
    //      - -> {0xFFFFFFFF, 0xFFFFAAXX}
    //     
    for (DWORD i = 0; i < NumDWords; i++)
        {
        DWORD data = lpBuffer[i];
        WORD mid;

        if (ReadRawData)
            Module[MODULE_INDEX_RAW].pvt_CratePut(data);

        int ModuleSlot = (data >> 8) & 0xF;

        /*    */
        if (special_code && ((data & 0xFFFFFF00) == 0xFFFFAA00))
            {
            if (!ProcessCrateError(data))
                {
                LOGWRITE(Format("%s: crate sent signal %0.2X. Cannot read crate registers",
                    ARRAYOFCONST((name, (int)data & 0xFF))),
                    LOGTASK_CRATE_READ, LOGL_ERROR);
                sflags |= CRATE_FLAG_HW_ERROR;
                result = false;
                }
            continue; //   switch
            }
        special_code = false;

        switch (data & LTR010_CMD_DATA_MASK)
            {
            /*    (   LTR030) */
            case LTR010USERDATA:
                ModuleSlot = MODULE_INDEX_USER;
                /* no break! */
            /*   */
            case LTR010DATA:
                //     
                Module[ModuleSlot].pvt_CratePut(data);
                break;

            case LTR010CMD:
                //   
                if (!Module[ModuleSlot].Present())
                    break;
                mid = Module[ModuleSlot].mid;

                //     41, 42, 43- ,    LTR021
                if  (((data & 0xCFFFC0C0) == LTR010CMD) &&
                     ((mid == LTR_MID(41)) || (mid == LTR_MID(42)) ||
                     (mid == LTR_MID(43)) || (mid == LTR_MID(21)))
                    )
                    {
                    ProcessSyncMark(data >> 16);
                    break;
                    }

                if ((mid == LTR_MID(34)) && ((data & 0x80E0) == 0x80C0))
                    {   //    34- 
                    //      ,      
                    Module[ModuleSlot].pvt_CrateUpdateLTR34(
                        1024, ((data >> 16) & 0xFFFF) | ((data & 0x11) << 16));
                    break;
                    }

                // LTR34 echo data
                if ((mid == LTR_MID(34)) && ((data & 0x80E0) == 0x80E0))
                    {
                    Module[ModuleSlot].pvt_CrateUpdateLTR34((data & 0x11) << 16);
                    }

                //    
                //  LTR11 -  -  -     ,
                //        ,     
                //    
#if 1
                if ((data & ((mid == LTR_MID(11)) ? 0xF0FF : 0xC0C0)) == LTR010CMD_RESET)
#else
                // , ,    -- 0xC0C0  0xF0FF
                if (((data & 0xC0C0) == LTR010CMD_RESET) ||
                    ((mid == LTR_MID(11)) && ((data & 0xF0FF) == LTR010CMD_RESET)))
#endif
                    {
                    WORD new_mid = data >> 16;
                    bool bad = false;
                    if (((mid & 0xFF) != ((mid >> 8) & 0xFF))
                        && ((data & 0xF0FF) != LTR010CMD_RESET))
                        { //   ( LTR11)
                        bad = true;
                        }
                    else if ((new_mid & 0xFF) != ((new_mid >> 8) & 0xFF))
                        {
                        bad = true;
                        }

                    if (bad)
                        {
                        LOGWRITE(Format("%s: bad module (slot = %d, mid = %0.4X, data = %0.8X)",
                            ARRAYOFCONST((name, (int)ModuleSlot + 1, (int)mid, data))),
                            LOGTASK_CRATE_READ, LOGL_ERROR);
                        }
                    else if (mid != new_mid)
                        {
                        InitLockedModule(ModuleSlot, new_mid, module_init_normal);
                        refMQ_Display.put(LTR_DISP_crate_list_updated);
                        }
                    } /* reset */

                Module[ModuleSlot].pvt_CratePut(data);
                break;

            case LTR010CRATE_CMD:
                {
                //    
                /*    -    :
                 *  -       ( ..   )
                 *    :   FFFFFFFF,  FFFFAAxx
                 *      - 0xAA & 0xC0 == 0x80      
                 *    (  );    special_code
                 *  -    
                 *    : x000FF10,  28 = start,  29 = 1sec
                 *
                 *       16  .
                 */
                WORD CmdCode = data & 0xFFFF;
                switch (CmdCode)
                    {
                    case 0xFF10: //   
                        ProcessSyncMark(data >> 16);
                        break;
                    case 0xFFFF: //  
                        special_code = true;
                        break;
                    default:
                        LOGWRITE(Format("%s: unknown crate command code (data = %0.8X)",
                            ARRAYOFCONST((name, data))),
                            LOGTASK_CRATE_READ, LOGL_ERROR);
                        break;
                    } /* switch CmdCode */
                }
                break;

            default:
                LOGWRITE(Format("%s: bad crate data (invalid bit combination) data = %0.8X ",
                    ARRAYOFCONST((name, data))),
                    LOGTASK_CRATE_READ, LOGL_ERROR);
                break;
            } /* switch (data & LTR010_CMD_DATA_MASK) */
        } /* for i */

    UnlockAllModules();
    return result;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
#define NUM_MIX_WORDS_FOR_ONE_FAST_MODULE   20
DWORD TCrate::PrepareDataForCrate(LPDWORD lpBuffer, DWORD MaxDWords, DWORD &SkipCount)
    {
    DWORD size = 0;

    if (!MaxDWords || !lpBuffer)
        return 0;

#if DEBUG_PROFILER
    Prf::Add _(0);
#endif

    LockAllModules();

    //   STOP+RESET+STOP ,  
    for (int i = 0; i < MODULE_MAX; i++)
        {
        if (size + 3 > MaxDWords) break;
        if (Module[i].pvt_GetAndClearInitFlag())
            { //   
            DWORD slot_field = i << 8;
            lpBuffer[size + 0] = lpBuffer[size + 2] = LTR010CMD_STOP | slot_field;
            lpBuffer[size + 1] = (LTR010CMD_RESET | 0x01010000) | slot_field;
            size += 3;
            LOGWRITE(Format("%s: sending STOP+RESET+STOP to slot %d",
                ARRAYOFCONST((name, (i + 1)))),
                LOGTASK_CRATE_WRITE, LOGL_INFO);
            }
        }

    //         
    //    
    DWORD fast_data_mask = 0; //        
    bool have_data = false; // ,        
    for (int i = 0, bit_i = 1; i < MODULE_MAX; i++, bit_i <<= 1)
        {
        if (Module[i].pvt_CrateDataReady())
            {
            have_data = true;
            if ((baud_rate & bit_i))
                fast_data_mask |= bit_i;
            }
        }

    //    &&!IsStopped (,   ),
    //     .      .
    while (have_data && (size < MaxDWords))
        {
        have_data = false;
        DWORD next_skip_count = (SkipCount) ? SkipCount - 1 : 0;

        //      ...
        if (fast_data_mask)
            { //       
            if (!SkipCount)
                next_skip_count = NUM_MIX_WORDS_FOR_ONE_FAST_MODULE - 1;
            }
        else
            { //        
            SkipCount = next_skip_count = 0;
            }

        _P;
        //          
        for (int i = 0, bit_i = 1; i < MODULE_MAX; i++, bit_i <<= 1)
            {
            if (SkipCount && !(fast_data_mask & bit_i))
                continue;
            if (!Module[i].pvt_CrateDataReady())
                {
                fast_data_mask &= ~bit_i;
                continue;
                }
            lpBuffer[size++] = Module[i].pvt_CrateGet();
            have_data = true;
            if (size >= MaxDWords) break;
            }
        _R;

        //  ,  
        SkipCount = next_skip_count;
        } /* while size */

    UnlockAllModules();
    DEBUGDUMP(lpBuffer, size * sizeof(DWORD), Format("to crate %s", ARRAYOFCONST((name))).c_str());
    return size;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
bool TCrateUSB::GetArray(DWORD mode, DWORD address, BYTE *data, DWORD size)
    {
    bool rv;
    switch (mode)
        {
        case CM_APP:
            rv = ltr_USB->Application->GetArray(address, data, size);
            break;
        case CM_BOOT:
            rv = ltr_USB->Bootloader->GetArray(address, data, size);
            break;
        case CM_AUTO:
        default:
            rv = (ltr_USB->application_active)
                ? ltr_USB->Application->GetArray(address, data, size)
                : ltr_USB->Bootloader->GetArray(address, data, size);
            break;
        }
    if (rv)
        {
        unsigned int retries = ltr_USB->GetLastIoctlRetries();
        if (retries > 0)
            LOGWRITE(Format("%s: GetArray() retried %d times", ARRAYOFCONST((name, retries))),
                LOGTASK_CRATE_CTL, LOGL_DEBUG3);
        }
    else
        {
        LOGWRITE(Format("%s: GetArray() failed.", ARRAYOFCONST((name))),
            LOGTASK_CRATE_CTL, LOGL_DEBUG2);
        }
    return rv;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
bool TCrateIP::GetArray(DWORD mode, DWORD address, BYTE *data, DWORD size)
    {
    if (mode == CM_BOOT)
        return false;
    bool rv = ltr_ip->GetArray(address, data, size);
    if (!rv)
        LOGWRITE(Format("%s: GetArray() failed: %s",
            ARRAYOFCONST((name, GetErrorMessage()))),
            LOGTASK_CRATE_CTL, LOGL_DEBUG2);
    return rv;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
bool TCrateUSB::PutArray(DWORD mode, DWORD address, BYTE *data, DWORD size)
    {
    bool rv;
    switch (mode)
        {
        case CM_APP:
            rv = ltr_USB->Application->PutArray(address, data, size);
            break;
        case CM_BOOT:
            rv = ltr_USB->Bootloader->PutArray(address, data, size);
            break;
        case CM_AUTO:
        default:
            rv = (ltr_USB->application_active)
                ? ltr_USB->Application->PutArray(address, data, size)
                : ltr_USB->Bootloader->PutArray(address, data, size);
            break;
        }
    if (rv)
        {
        unsigned int retries = ltr_USB->GetLastIoctlRetries();
        if (retries > 0)
            LOGWRITE(Format("%s: PutArray() retried %d times", ARRAYOFCONST((name, retries))),
                LOGTASK_CRATE_CTL, LOGL_DEBUG3);
        }
    else
        {
        LOGWRITE(Format("%s: PutArray() failed.", ARRAYOFCONST((name))),
            LOGTASK_CRATE_CTL, LOGL_DEBUG2);
        }
    return rv;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
bool TCrateIP::PutArray(DWORD mode, DWORD address, BYTE *data, DWORD size)
    {
    if (mode == CM_BOOT)
        return false;
    bool rv = ltr_ip->PutArray(address, data, size);
    if (!rv)
        LOGWRITE(Format("%s: PutArray() failed: %s",
            ARRAYOFCONST((name, GetErrorMessage()))),
            LOGTASK_CRATE_CTL, LOGL_DEBUG2);
    return rv;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
bool TCrateUSB::GetModuleDescription(DWORD mode, TDESCRIPTION_LTR010 *description)
    {
    bool rv;
    switch (mode)
        {
        case CM_APP:
            rv = ltr_USB->Application->GetModuleDescription(description);
            break;
        case CM_BOOT:
            rv = ltr_USB->Bootloader->GetModuleDescription(description);
            break;
        case CM_AUTO:
        default:
            rv = (ltr_USB->application_active)
                ? ltr_USB->Application->GetModuleDescription(description)
                : ltr_USB->Bootloader->GetModuleDescription(description);
            break;
        }
#if 0 // this diagnostic is done on upper level
    if (rv)
        {
        unsigned int retries = ltr_USB->GetLastIoctlRetries();
        if (retries > 0)
            LOGWRITE(Format("%s: GetModuleDescription() retried %d times", ARRAYOFCONST((name, retries))),
                LOGTASK_CRATE_CTL, LOGL_DEBUG3);
        }
    else
        {
        LOGWRITE(Format("%s: GetModuleDescription() failed.", ARRAYOFCONST((name))),
            LOGTASK_CRATE_CTL, LOGL_DEBUG2);
        }
#endif
    return rv;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
bool TCrateIP::GetModuleDescription(DWORD mode, TDESCRIPTION_LTR010 *description)
    {
    if (mode == CM_BOOT)
        return false;
    bool rv = ltr_ip->GetModuleDescription(description);
    if (!rv)
        LOGWRITE(Format("%s: GetModuleDescription() failed: %s",
            ARRAYOFCONST((name, GetErrorMessage()))),
            LOGTASK_CRATE_CTL, LOGL_DEBUG2);
    return rv;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
bool TCrateUSB::SetModuleDescription(DWORD mode, TDESCRIPTION_LTR010 *description)
    {
    switch (mode)
        {
        case CM_APP:
            return false;
        case CM_BOOT:
            return ltr_USB->Bootloader->SetModuleDescription(description);
        case CM_AUTO:
        default:
            return (ltr_USB->application_active)
                ? false
                : ltr_USB->Bootloader->SetModuleDescription(description);
        }
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
bool TCrateIP::SetModuleDescription(DWORD mode, TDESCRIPTION_LTR010 *description)
    {
    return false; //   ,    IP  
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
bool TCrateUSB::CallApplication(DWORD address)
    {
    BYTE tempdata[16];

    if (!ltr_USB->Bootloader->CallApplication(address))
        return false;
    /* ,   :  application  bootloader */
    if (ltr_USB->Bootloader->GetArray(0x0 | SEL_AVR_PM, tempdata, 16))
        {
        ltr_USB->application_active = false;
        }
    else if (ltr_USB->Application->GetArray(0x0 | SEL_AVR_PM, tempdata, 16))
        {
        ltr_USB->application_active = true;
        }
    return true;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
bool TCrateIP::CallApplication(DWORD address)
    {
    return false; //   ,    IP  
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
void TCrateUSB::CancelIO(void)
    {
    ltr_USB->CancelIO();
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
void TCrateIP::CancelIO(void)
    {
    // Do nothing
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
int TCrateUSB::PingAndGetFailCount(DWORD timeout)
    {
    return 0;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
int TCrateIP::PingAndGetFailCount(DWORD timeout)
    {
#if USE_EXTENDED_PING
    DWORD diag[3];
    DWORD rx_bytes;
    if (ltr_ip->ExecCommand(LTR_IP_CMD_NOP, 0,
                NULL, 0, (BYTE*)diag, sizeof(diag),
                NULL, &rx_bytes, timeout))
        {
        if (rx_bytes >= sizeof(diag))
            LOGWRITE(Format("%s: Ping() -> diag: %u, %u, %u",
                ARRAYOFCONST((name, diag[0], diag[1], diag[2]))),
                LOGTASK_CRATE_CTL, LOGL_DEBUG3);
#else
    if (ltr_ip->Ping(timeout))
        {
#endif
        ping_fail_count = 0;
        }
    else
        {
        if (ping_fail_count < INT_MAX)
            ping_fail_count++;
        LOGWRITE(Format("%s: Ping() failed: %s",
            ARRAYOFCONST((name, GetErrorMessage()))),
            LOGTASK_CRATE_CTL, LOGL_DEBUG2);
        }
    return ping_fail_count;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
AnsiString TCrateIP::GetErrorMessage(void)
    {
    DWORD err = ltr_ip->GetLastErrorCode();
    int subcode = err & 0xFFFFFF;
    AnsiString msg;
    if (subcode & 0x800000)
        subcode |= 0xFF000000;
    switch (err & 0xFF000000)
        {
        case LTR_IPERR_OK:              msg = "No error";                   break;
        case LTR_IPERR_INVALID_PARAM:   msg = "Invalid parameter in call";  break;
        case LTR_IPERR_CONNECT:         msg = "Connect failed";             break;
        case LTR_IPERR_ABORT:           msg = "Aborted";                    break;
        case LTR_IPERR_TIMEOUT_W:       msg = "Write timeout";              break;
        case LTR_IPERR_TIMEOUT_R:       msg = "Read timeout";               break;
        case LTR_IPERR_CONN_LOST:       msg = "Connection lost";            break;
        case LTR_IPERR_READ:            msg = "Read error";                 break;
        case LTR_IPERR_WRITE:           msg = "Write error";                break;
        case LTR_IPERR_BAD_SIG:         msg = "Bad signature in reply";     break;
        case LTR_IPERR_BAD_REPLYLEN:    msg = "Invalid reply length";       break;
        case LTR_IPERR_CMD_FAILED:      msg = "Command failed";             break;
        case LTR_IPERR_NOT_CONNECTED:   msg = "Not connected";              break;
        case LTR_IPERR_INTERNAL:        msg = "Internal error";             break;
        default:                        msg = "Unknwon error";              break;
        }
    if (subcode != 0)
        msg = msg + " (" + subcode + ")";
    return msg;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
static AnsiString f_GetFPGAErrorMessage(INT fpga_err_code)
    {
    switch (fpga_err_code)
        {
        case LTR_FPGA_ERR_BADPARAM:     return "Invalid LoadFPGA parameters";
        case LTR_FPGA_ERR_FILEIO:       return "Error reading TTF file";
        case LTR_FPGA_ERR_INITDMAC:     return "InitDMAC failed";
        case LTR_FPGA_ERR_INITDMAC0:    return "InitDMAC(0) failed";
        case LTR_FPGA_ERR_INITDMAC7:    return "InitDMAC(7) failed";
        case LTR_FPGA_ERR_RESETFPGA:    return "ResetFPGA failed";
        case LTR_FPGA_ERR_INITFPGA:     return "InitFPGA failed";
        case LTR_FPGA_ERR_PUTARRAY:     return "PutArray failed";
        case LTR_FPGA_ERR_READBULK:     return "ReadBulk failed (unit in TCP/IP mode?)";
        case LTR_FPGA_ERR_BADREPLY:     return "ReadBulk returned invalid data";
        case LTR_FPGA_ERR_NOREPLY:      return "ReadBulk returned 0 bytes";
        default:                        return "Unknown FPGA error";
        }
    }
/*------------------------------------------------------------------------------------------------*/

